#!/usr/bin/perl

# Unix SMB/CIFS implementation.
# Test suite for the smbcacls binary
# Copyright (C) Aurélien Aptel 2013
# Copyright (C) David Disseldorp 2013
# Copyright (C) Noel Power 2013

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# TODO - split all common test_smbcacls.pl/test_smbclient_tarmode.pl code into
# a separate file.

=head1 NAME

C<test_smbcacls.pl> - Test for smbcacls

=cut

use v5.10;
use strict;
use warnings;

use Archive::Tar;
use Data::Dumper;
use Digest::MD5 qw/md5_hex/;
use File::Path qw/make_path remove_tree/;
use File::Spec;
use File::Temp;
use Getopt::Long;
use Pod::Usage;
use Term::ANSIColor;

sub d {print Dumper @_;}

# DEFAULTS
# 'our' to make them available in the File package
our $USER      = '';
our $PW        = '';
our $HOST      = '';
our $SHARE     = '';
our $DIR       = 'tar_test_dir';
our $LOCALPATH = '';
our $TMP       = File::Temp->newdir();
our $CLI_BIN   = 'smbclient';
our $CACLS_BIN = 'smbcacls';
our $SUBUNIT   = 0;

my $SELECTED_TEST = '';
my $LIST_TEST = 0;

my @SMBARGS   = ();

our $DEBUG = 1;
our $VERBOSE = 1;
my $MAN   = 0;
my $HELP  = 0;
my $CLEAN = 0;

# all tests
my @TESTS = (
    ['simple, single file set ACL', \&test_simple_single_set],
    ['simple, single file modify ACL', \&test_simple_single_mod],
    ['simple, single file delete ACL', \&test_simple_single_del],
    ['simple, single file add ACL', \&test_simple_single_add],
    ['simple, single object inherit add', \&test_simple_oi_add],
    ['simple, single object inherit delete', \&test_simple_oi_delete],
    ['simple, single object inherit modify', \&test_simple_oi_modify],
    ['simple, single container inherit add', \&test_simple_ci_add],
    ['simple, single container inherit delete', \&test_simple_ci_delete],
    ['simple, single container inherit modify', \&test_simple_ci_modify],
    ['simple, single container/object inherit add', \&test_simple_cioi_add],
    ['simple, single container/object inherit delete', \&test_simple_cioi_delete],
    ['simple, single container/object inherit modify', \&test_simple_cioi_modify],
    ['simple, single inherit set(fail test)', \&test_simple_set_fail],
    ['simple, single container/object inherit set', \&test_simple_oici_set],
    ['simple, single container inherit set', \&test_simple_ci_set],
    ['simple, single container/inherit/np add', \&test_simple_cioinp_add],
    ['simple, single object inherit/np add', \&test_simple_oinp_add],
    ['simple, single container inherit/np add', \&test_simple_cinp_add],
    ['simple, single container/inherit/np delete', \&test_simple_cioinp_delete],
    ['simple, single object inherit/np delete', \&test_simple_oinp_delete],
    ['simple, single container inherit/np delete', \&test_simple_cinp_delete],
    ['simple, single container/object inherit inhibit', \&test_simple_cioi_inhibit],
);

=head1 SYNOPSIS

 test_smbcacls.pl [options] -- [smbclient options]

 Options:
    -h, --help    brief help message
    --man         full documentation

  Environment:
    -u, --user      USER
    -p, --password  PW
    -n, --name      HOST	(required)
    -s, --share     SHARE	(required)
    -d, --dir       PATH
        sub-path to use on the share

    -l, --local-path  PATH	(required)
        path to the root of the samba share on the machine.

    --cli_bin       BIN
        path to the smbclient binary to use

    --cacls_bin     BIN
        path to the smbcacls binary to use

  Test:
    --list
       list tests

    --test N
    --test A-B
    --test A,B,D-F
       only run certain tests (accept list and intervals of numbers)

    -v, --verbose
       be more verbose

    --debug
       print command and their output (also set -v)

    --subunit
       print output in subunit format

    --clean
       clean out DIR path between tests

=cut

GetOptions('u|user=s'       => \$USER,
           'p|password=s'   => \$PW,
           'n|name=s'       => \$HOST,
           's|share=s'      => \$SHARE,
           'd|dir=s'        => \$DIR,
           'l|local-path=s' => \$LOCALPATH,
           'cli_bin=s'    => \$CLI_BIN,
           'cacls_bin=s'  => \$CACLS_BIN,

           'test=s'         => \$SELECTED_TEST,
           'list'           => \$LIST_TEST,

           'clean'          => \$CLEAN,
           'subunit'        => \$SUBUNIT,
           'debug'          => \$DEBUG,
           'v|verbose'      => \$VERBOSE,
           'h|help'         => \$HELP,
           'man'            => \$MAN) or pod2usage(2);

pod2usage(0) if $HELP;
pod2usage(-exitval => 0, -verbose => 2) if $MAN;
list_test(), exit 0 if $LIST_TEST;
pod2usage(1) unless $HOST;
pod2usage(1) unless $SHARE;
pod2usage(1) unless $LOCALPATH;
pod2usage(1) unless $USER;
pod2usage(1) unless $PW;

push @SMBARGS, '-U'.$USER.'%'.$PW;

# remaining arguments are passed to smbclient
push @SMBARGS, @ARGV;

#####

# SANITIZATION

# remove all final slashes from input paths
$LOCALPATH =~ s{[/\\]+$}{}g;
$SHARE =~ s{[/\\]+$}{}g;
$HOST =~ s{[/\\]+$}{}g;
$DIR =~ s{[/\\]+$}{}g;
# remove ./ prefix from DIR
$DIR =~ s{^[\./\\]+}{}g;

if (!-d $LOCALPATH) {
    die "Local path '$LOCALPATH' is not a directory.\n";
}

if ($CLEAN) {
    # clean the whole root first
    remove_tree($LOCALPATH, { keep_root => 1 });
}

if ($DEBUG) {
    $VERBOSE = 1;
}

#####

# RUN TESTS

my @selection = parse_test_string($SELECTED_TEST);

if ($SELECTED_TEST eq '') {
    run_test(@TESTS);
} elsif (@selection > 0) {
    run_test(@selection);
} else {
    die "Test selection '$SELECTED_TEST' is invalid\n";
}

#################################

=head1 DOCUMENTATION

=head2 Defining a test

=over

=item * Create a function C<test_yourtest>

=item * Use the File module, documented below

=item * Use C<smb_tar>, C<smb_client>, C<smb_cacls>, C<check_tar> or C<check_remote>

=item * Return number of errors

=item * Add function to C<@TESTS>

=back

The function must be placed in the C<@TESTS> list along with a short
description and optional arguments.

=cut


# test smbcacls '--set' attempts to overwrite the ACL for the file
#
# before:
#
#  +-tar_test_dir/
#    +-file.1            (I)(F)
#
# after/expected:
#
#  +-tar_test_dir/
#    +-file.1            (F)
#

sub test_simple_single_set
{
	my @files;
	my $acl_str = "ACL:$USER:ALLOWED/0x0/FULL";
	my $ace;
	my $ret;
	my $out;

	my $f = File->new_local("$TMP/file-1");
	$f->put_remote("file-1");

	($out, $ret) = smb_cacls('--set', $acl_str, $f->smb_remotepath);
	if ($ret != 0 ) {
		return 1;
	}
	# only a single ACE string in the ACL
	$ace = ace_parse_str($acl_str);
	$ret = file_ace_check($f->smb_remotepath, $ace);

	$f->del_remote(1);

	return 0;
}


# test smbcacls '--modify' attempts to modify the ACL for the file
# (note: first part of the test 'set' ACL to (F) then attempts to modify
# before:
#
#  +-tar_test_dir/
#    +-file.1            (F)
#
# after/expected:
#
#  +-tar_test_dir/
#    +-file.1            (READ)
#
sub test_simple_single_mod
{
	my @files;
	my $acl_str = "ACL:$USER:ALLOWED/0x0/FULL";
	my $ace;
	my $ret;
	my $out;

	my $f = File->new_local("$TMP/file-1");
	$f->put_remote("file-1");

	($out, $ret) = smb_cacls('--set', $acl_str, $f->smb_remotepath);
	if ($ret != 0 ) {
		return 1;
	}
	# only a single ACE string in the ACL
	$ace = ace_parse_str($acl_str);
	$ret = file_ace_check($f->smb_remotepath, $ace);
	if ($ret != 0) {
		return 1;
	}

	# overwrite existing entry
	$acl_str = "ACL:$USER:ALLOWED/0x0/READ";
	($out, $ret) = smb_cacls('--modify', $acl_str, $f->smb_remotepath);
	if ($ret != 0) {
		return 1;
	}
	$ace = ace_parse_str($acl_str);
	$ret = file_ace_check($f->smb_remotepath, $ace);

	$f->del_remote(1);

	return 0;
}

# test smbcacls '--delete' attempts to delete the ACL for the file
# (note: first part of the test 'set' ACL to (F) then attempts to delete
# before:
#
#  +-tar_test_dir/
#    +-file.1            (F)
#
# after/expected:
#
#  +-tar_test_dir/
#    +-file.1            (none) - meaning no (F) ACL for this user
#
sub test_simple_single_del
{
	my @files;
	my $acl_str = "ACL:$USER:ALLOWED/0x0/FULL";
	my $ace;
	my $ret;
	my $out;
	my $f = File->new_local("$TMP/file-1");
	$f->put_remote("file-1");

	smb_cacls('--set', $acl_str, $f->smb_remotepath);

	# only a single ACE string in the ACL
	$ace = ace_parse_str($acl_str);
	$ret = file_ace_check($f->smb_remotepath, $ace);
	if ($ret != 0) {
		return 1;
	}

	($out, $ret) = smb_cacls('--delete', $acl_str, $f->smb_remotepath);
	if ($ret != 0 ) {
		return 1;
	}
	$ace = ace_parse_str($acl_str);
	$ret = file_ace_check($f->smb_remotepath, $ace);
	if ($ret == 0) {
		say "got ace where not expecting";
		return 1;
	}

	$f->del_remote(1);

	return 0;
}

# test smbcacls '--add' attempts to add the ACL for the file
# (note: first part of the test 'set' ACL to (F) then attempts to add new ACL
# before:
#
#  +-tar_test_dir/
#    +-file.1            (F)
#
# after/expected:
#
#  +-tar_test_dir/       
#    +-file.1            (F,READ)
#
sub test_simple_single_add
{
	my @files;
	my $acl_str = "ACL:$USER:ALLOWED/0x0/FULL";
	my $dny_acl_str = "ACL:$USER:DENIED/0x0/READ";
	my $ace;
	my $ret;
	my $out;

	my $f = File->new_local("$TMP/file-1");
	$f->put_remote("file-1");

	($out, $ret) = smb_cacls('--set', $acl_str, $f->smb_remotepath);
	if ($ret != 0 ) {
		return 1;
	}
	# only a single ACE string in the ACL
	$ace = ace_parse_str($acl_str);
	$ret = file_ace_check($f->smb_remotepath, $ace);
	if ($ret != 0) {
		say "missing ace";
		return 1;
	}

	($out, $ret) = smb_cacls('--add', $dny_acl_str, $f->smb_remotepath);
	if ($ret != 0 ) {
		return 1;
	}
	$ace = ace_parse_str($dny_acl_str);
	$ret = file_ace_check($f->smb_remotepath, $ace);
	if ($ret != 0) {
		say "missing ace";
		return 1;
	}

	$f->del_remote(1);

	return 0;
}


# test smbcacls '--propagate-inheritance --add' which attempts to add the ACL
# for the file and additionally use inheritance rules to propagate appropriate
# changes to children
#
# This test adds an ACL with (OI)(READ)
#
# before:
#
#  +-tar_test_dir/    (01)(CI)(I)(F)
#    +-oi_dir/        (OI)(CI)(I)(F)
#    | +-file.1            (I)(F)
#    | +-nested/      (OI)(CI)(I)(F)
#    |   +-file.2          (I)(F)
#
# after/expected:
#
#  +-tar_test_dir/    (01)(CI)(I)(F)
#    +-oi_dir/        (OI)(CI)(I)(F), (OI)(READ)
#    | +-file.1            (I)(F), (I)(READ)
#    | +-nested/      (OI)(CI)(I)(F), (OI)(IO)(I)(READ)
#    |   +-file.2          (I)(F), (I)(READ)
#

sub test_simple_oi_add
{
	my @files;
	my $dir_add_acl_str = "ACL:$USER:ALLOWED/OI/READ";
	my $obj_inherited_ace_str = "ACL:$USER:ALLOWED/I/READ";
	my $dir_inherited_ace_str = "ACL:$USER:ALLOWED/OI|IO|I/READ";
	my $dir_ace;
	my $child_file_ace;
	my $child_dir_ace;
	my $ret;
	my $out;

	my $f1 = File->new_local("$TMP/file-1");
	$f1->put_remote("oi_dir/file-1");
	my $f2 = File->new_local("$TMP/file-2");
	$f2->put_remote("oi_dir/nested/file-2");

	($out, $ret) = smb_cacls('--propagate-inheritance', '--add',
				 $dir_add_acl_str, $f1->smb_remotedir);
	if ($ret != 0 ) {
		return 1;
	}

	# check top level container 'oi_dir' has OI/READ
	$dir_ace = ace_parse_str($dir_add_acl_str);
	$ret = file_ace_check($f1->smb_remotedir, $dir_ace);
	if ($ret != 0) {
		say "missing ace";
		return 1;
	}
	# file 'oi_dir/file-1' should  have inherited I/READ
	$child_file_ace = ace_parse_str($obj_inherited_ace_str);
	$ret = file_ace_check($f1->smb_remotepath, $child_file_ace);

	if ($ret != 0) {
		say "missing ace";
		return 1;
	}

	# nested dir  'oi_dir/nested/' should have OI|IO/READ
	$child_dir_ace = ace_parse_str($dir_inherited_ace_str);
	$ret = file_ace_check($f2->smb_remotedir, $child_dir_ace);
	if ($ret != 0) {
		say "missing ace";
		return 1;
	}

	# nested file 'oi_dir/nested/file-2' should  have inherited I/READ
	$ret = file_ace_check($f2->smb_remotepath, $child_file_ace);
	if ($ret != 0) {
		say "got ace where not expecting";
		return 1;
	}

	$f1->del_remote();
	$f2->del_remote(1);

	return 0;
}
# test smbcacls '--propagate-inheritance --add' which attempts to add the ACL
# for the file and additionally use inheritance rules to propagate appropriate
# changes to children
#
# This test adds an ACL with (OI)(READ)
#
# before:
#
#  +-tar_test_dir/    (01)(CI)(I)(F)
#    +-oi_dir/        (OI)(CI)(I)(F), (OI)(IO)(READ)
#    | +-file.1            (I)(F), (I)(READ)
#    | +-nested/      (OI)(CI)(I)(F), (OI)(IO)(I)(READ)
#    |   +-file.2          (I)(F), (I)(READ)
#
# after/expected:
#
#  +-tar_test_dir/    (01)(CI)(I)(F)
#    +-oi_dir/        (OI)(CI)(I)(F)
#    | +-file.1            (I)(F)
#    | +-nested/      (OI)(CI)(I)(F)
#    |   +-file.2          (I)(F)
#
sub test_simple_oi_delete
{
	my @files;
	my $dir_acl_str = "ACL:$USER:ALLOWED/OI/READ";
	my $obj_inherited_ace_str = "ACL:$USER:ALLOWED/I/READ";
	my $dir_inherited_ace_str = "ACL:$USER:ALLOWED/OI|IO|I/READ";
	my $dir_ace;
	my $child_file_ace;
	my $child_dir_ace;
	my $ret;
	my $out;

	my $f1 = File->new_local("$TMP/file-1");
	$f1->put_remote("oi_dir/file-1");
	my $f2 = File->new_local("$TMP/file-2");
	$f2->put_remote("oi_dir/nested/file-2");

	# add flags on oi_dir
	($out, $ret) = smb_cacls('--add', $dir_acl_str, $f1->smb_remotedir);
	if ($ret != 0 ) {
		return 1;
	}
	# add flags on oi_dir/nested
	($out, $ret) = smb_cacls('--add', $dir_inherited_ace_str,
				 $f2->smb_remotedir);
	if ($ret != 0 ) {
		return 1;
	}
	# add flags on oi_dir/file-1
	($out, $ret) = 	smb_cacls('--add', $obj_inherited_ace_str,
		$f1->smb_remotepath);
	if ($ret != 0 ) {
		return 1;
	}
	# add flags on oi_dir/nested/file-2
	($out, $ret) = smb_cacls('--add', $obj_inherited_ace_str,
				 $f2->smb_remotepath);
	if ($ret != 0 ) {
		return 1;
	}
	($out, $ret) = smb_cacls('--propagate-inheritance', '--delete',
				 $dir_acl_str,
		  $f1->smb_remotedir);
	if ($ret != 0 ) {
		return 1;
	}

	# check top level container 'oi_dir' no longer has OI/READ
	$dir_ace = ace_parse_str($dir_acl_str);
	$ret = file_ace_check($f1->smb_remotedir, $dir_ace);
	if ($ret == 0) {
		say "unexpected ace";
		return 1;
	}
	# file 'oi_dir/file-1' should  no longer have inherited I/READ
	$child_file_ace = ace_parse_str($obj_inherited_ace_str);
	$ret = file_ace_check($f1->smb_remotepath, $child_file_ace);

	if ($ret == 0) {
		say "unexpected ace";
		return 1;
	}

	# nested dir  'oi_dir/nested/' should no longer have OI|IO/READ
	$child_dir_ace = ace_parse_str($dir_inherited_ace_str);
	$ret = file_ace_check($f2->smb_remotedir, $child_dir_ace);
	if ($ret == 0) {
		say "unexpected ace";
		return 1;
	}

	# nested file 'oi_dir/nested/file-2' should no longer have inherited I/READ
	$ret = file_ace_check($f2->smb_remotepath, $child_file_ace);
	if ($ret == 0) {
		say "got ace where not expecting";
		return 1;
	}
	$f1->del_remote();
	$f2->del_remote(1);

	return 0;
}

# test smbcacls '--propagate-inheritance --add' which attempts to add the ACL
# for the file and additionally use inheritance rules to propagate appropriate
# changes to children
#
# This test adds an ACL with (CI)(READ)
#
# before:
#
#  +-tar_test_dir/    (01)(CI)(I)(F)
#    +-oi_dir/        (OI)(CI)(I)(F)
#    | +-file.1            (I)(F)
#    | +-nested/      (OI)(CI)(I)(F)
#    |   +-file.2          (I)(F)
#
# after/expected:
#
#  +-tar_test_dir/    (01)(CI)(I)(F)
#    +-oi_dir/        (OI)(CI)(I)(F), (CI)(READ)
#    | +-file.1            (I)(F)
#    | +-nested/      (OI)(CI)(I)(F), (CI)((I)(READ)
#    |   +-file.2          (I)(F)
sub test_simple_ci_add
{
	my @files;
	my $dir_add_acl_str = "ACL:$USER:ALLOWED/CI/READ";
	my $file_inherited_ace_str = "ACL:$USER:ALLOWED/I/READ";
	my $dir_inherited_ace_str = "ACL:$USER:ALLOWED/CI|I/READ";
	my $dir_ace;
	my $child_file_ace;
	my $child_dir_ace;
	my $ret;
	my $out;

	my $f1 = File->new_local("$TMP/file-1");
	$f1->put_remote("oi_dir/file-1");
	my $f2 = File->new_local("$TMP/file-2");
	$f2->put_remote("oi_dir/nested/file-2");
	($out, $ret) = smb_cacls('--propagate-inheritance', '--add',
				 $dir_add_acl_str, $f1->smb_remotedir);
	if ($ret != 0 ) {
		return 1;
	}
	# check top level container 'oi_dir' has CI/READ
	$dir_ace = ace_parse_str($dir_add_acl_str);
	$ret = file_ace_check($f1->smb_remotedir, $dir_ace);
	if ($ret != 0) {
		say "missing ace";
		return 1;
	}


	$child_file_ace = ace_parse_str($file_inherited_ace_str);
	# nested file 'oi_dir/file-1' should NOT have inherited I/READ
	$ret = file_ace_check($f1->smb_remotepath, $child_file_ace);
	if ($ret == 0) {
		say "got ace where not expecting";
		return 1;
	}

	# nested dir  'oi_dir/nested/' should have CI|I|READ
	$child_dir_ace = ace_parse_str($dir_inherited_ace_str);
	$ret = file_ace_check($f2->smb_remotedir, $child_dir_ace);
	if ($ret != 0) {
		return 1;
	}

	$f1->del_remote();
	$f2->del_remote(1);

	return 0;
}

# test smbcacls '--propagate-inheritance --add' which attempts to add the ACL
# for the file and additionally use inheritance rules to propagate appropriate
# changes to children
#
# This test delete an ACL with (CI)(READ)
#
# before:
#
#  +-tar_test_dir/    (01)(CI)(I)(F)
#    +-oi_dir/        (OI)(CI)(I)(F), (CI)(READ)
#    | +-file.1            (I)(F)
#    | +-nested/      (OI)(CI)(I)(F), (CI)((I)(READ)
#    |   +-file.2          (I)(F)
#
# after/expected:
#
#  +-tar_test_dir/    (01)(CI)(I)(F)
#    +-oi_dir/        (OI)(CI)(I)(F)
#    | +-file.1            (I)(F)
#    | +-nested/      (OI)(CI)(I)(F)
#    |   +-file.2          (I)(F)
sub test_simple_ci_delete
{
	my @files;
	my $dir_acl_str = "ACL:$USER:ALLOWED/CI/READ";
	my $file_inherited_ace_str = "ACL:$USER:ALLOWED/I/READ";
	my $dir_inherited_ace_str = "ACL:$USER:ALLOWED/CI|I/READ";
	my $dir_ace;
	my $child_file_ace;
	my $child_dir_ace;
	my $ret;
	my $out;

	my $f1 = File->new_local("$TMP/file-1");
	$f1->put_remote("oi_dir/file-1");
	my $f2 = File->new_local("$TMP/file-2");
	$f2->put_remote("oi_dir/nested/file-2");

	# add flags on oi_dir
	($out, $ret) = smb_cacls('--add', $dir_acl_str, $f1->smb_remotedir);
	if ($ret != 0 ) {
		return 1;
	}
	# add flags on oi_dir/nested
	($out, $ret) = smb_cacls('--add', $dir_inherited_ace_str,
				 $f2->smb_remotedir);
	if ($ret != 0 ) {
		return 1;
	}
	# make sure no (I|RX) flags on oi_dir/file-1
	($out, $ret) = smb_cacls('--delete', $file_inherited_ace_str,
				 $f1->smb_remotepath);
	if ($ret != 0 ) {
		return 1;
	}
	# make sure no (I|RX) flags on oi_dir/nested/file-2
	($out, $ret) = smb_cacls('--delete', $file_inherited_ace_str,
				 $f2->smb_remotepath);
	if ($ret != 0 ) {
		return 1;
	}
	($out, $ret) = smb_cacls('--propagate-inheritance', '--delete',
				 $dir_acl_str, $f1->smb_remotedir);

	if ($ret != 0 ) {
		return 1;
	}
	# check top level container 'oi_dir' no longer has CI/READ
	$dir_ace = ace_parse_str($dir_acl_str);
	$ret = file_ace_check($f1->smb_remotedir, $dir_ace);

	if ($ret == 0) {
		say "unexpectex ace";
		return 1;
	}

	$child_file_ace = ace_parse_str($file_inherited_ace_str);
	# nested file 'oi_dir/file-1' should NOT have inherited I/READ
	$ret = file_ace_check($f1->smb_remotepath, $child_file_ace);
	if ($ret == 0) {
		say "got ace where not expecting";
		return 1;
	}

	# nested dir  'oi_dir/nested/' should no longer have CI|I|READ
	$child_dir_ace = ace_parse_str($dir_inherited_ace_str);
	$ret = file_ace_check($f2->smb_remotedir, $child_dir_ace);
	if ($ret == 0) {
		say "unexpected ace";
		return 1;
	}

	$f1->del_remote();
	$f2->del_remote(1);

	return 0;
}

# test smbcacls '--propagate-inheritance --add' which attempts to add the ACL
# for the file and additionally use inheritance rules to propagate appropriate
# changes to children
#
# This test adds an ACL with (CI)(OI)(READ)
#
# before:
#
#  +-tar_test_dir/    (01)(CI)(I)(F)
#    +-oi_dir/        (OI)(CI)(I)(F)
#    | +-file.1            (I)(F)
#    | +-nested/      (OI)(CI)(I)(F)
#    |   +-file.2          (I)(F)
#
# after/expected:
#
#  +-tar_test_dir/    (01)(CI)(I)(F)
#    +-oi_dir/        (OI)(CI)(I)(F), (CI)(OI)READ)
#    | +-file.1            (I)(F), (I)(READ)
#    | +-nested/      (OI)(CI)(I)(F), (CI)(OI)(I)(READ)
#    |   +-file.2          (I)(F), (I)(READ)
sub test_simple_cioi_add
{
	my @files;
	my $dir_add_acl_str = "ACL:$USER:ALLOWED/OI|CI/READ";
	my $file_inherited_ace_str = "ACL:$USER:ALLOWED/I/READ";
	my $dir_inherited_ace_str = "ACL:$USER:ALLOWED/OI|CI|I/READ";
	my $dir_ace;
	my $child_file_ace;
	my $child_dir_ace;
	my $ret;
	my $out;

	my $f1 = File->new_local("$TMP/file-1");
	$f1->put_remote("oi_dir/file-1");
	my $f2 = File->new_local("$TMP/file-2");
	$f2->put_remote("oi_dir/nested/file-2");

	($out, $ret) = smb_cacls('--propagate-inheritance', '--add',
				 $dir_add_acl_str, $f1->smb_remotedir);
	if ($ret != 0 ) {
		return 1;
	}
	# check top level container 'oi_dir' has OI|CI/READ
	$dir_ace = ace_parse_str($dir_add_acl_str);
	$ret = file_ace_check($f1->smb_remotedir, $dir_ace);
	if ($ret != 0) {
		say "missing expected ace";
		return 1;
	}


	$child_file_ace = ace_parse_str($file_inherited_ace_str);
	# nested file 'oi_dir/file-1' should have inherited I/READ
	$ret = file_ace_check($f1->smb_remotepath, $child_file_ace);
	if ($ret != 0) {
		say "missing expected ace";
		return 1;
	}

	# nested dir  'oi_dir/nested/' should have OI|CI|I|READ
	$child_dir_ace = ace_parse_str($dir_inherited_ace_str);
	$ret = file_ace_check($f2->smb_remotedir, $child_dir_ace);
	if ($ret != 0) {
		say "missing expected ace";
		return 1;
	}

	$f1->del_remote();
	$f2->del_remote(1);

	return 0;
}

# test smbcacls '--propagate-inheritance --delete' which attempts to delete the
# ACL for the file and additionally use inheritance rules to propagate
# appropriate changes to children
#
# This test deletes an ACL with (CI)(OI)(READ)
#
# before:
#
#  +-tar_test_dir/    (01)(CI)(I)(F)
#    +-oi_dir/        (OI)(CI)(I)(F), (CI)(OI)(READ)
#    | +-file.1            (I)(F), (I)(READ)
#    | +-nested/      (OI)(CI)(I)(F), (CI)(OI)(I)(READ)
#    |   +-file.2          (I)(F), (I)(READ)
#
# after/expected:
#
#  +-tar_test_dir/    (01)(CI)(I)(F)
#    +-oi_dir/        (OI)(CI)(I)(F)
#    | +-file.1            (I)(F)
#    | +-nested/      (OI)(CI)(I)(F)
#    |   +-file.2          (I)(F)
#

sub test_simple_cioi_delete
{
	my @files;
	my $dir_acl_str = "ACL:$USER:ALLOWED/OI|CI/READ";
	my $file_inherited_ace_str = "ACL:$USER:ALLOWED/I/READ";
	my $dir_inherited_ace_str = "ACL:$USER:ALLOWED/OI|CI|I/READ";
	my $dir_ace;
	my $child_file_ace;
	my $child_dir_ace;
	my $ret;
	my $out;

	my $f1 = File->new_local("$TMP/file-1");
	$f1->put_remote("oi_dir/file-1");
	my $f2 = File->new_local("$TMP/file-2");
	$f2->put_remote("oi_dir/nested/file-2");

	# add flags on oi_dir
	($out, $ret) = smb_cacls('--add', $dir_acl_str, $f1->smb_remotedir);
	if ($ret != 0) {
		return 1;
	}
	# add flags on oi_dir/nested
	($out, $ret) = smb_cacls('--add', $dir_inherited_ace_str,
				 $f2->smb_remotedir);
	if ($ret != 0) {
		return 1;
	}
	# add flags on oi_dir/file-1
	($out, $ret) = smb_cacls('--add', $file_inherited_ace_str,
				 $f1->smb_remotepath);
	if ($ret != 0) {
		return 1;
	}
	# add flags on oi_dir/nested/file-2
	($out, $ret) = smb_cacls('--add', $file_inherited_ace_str,
				 $f2->smb_remotepath);
	if ($ret != 0) {
		return 1;
	}

	($out, $ret) = smb_cacls('--propagate-inheritance', '--delete',
				 $dir_acl_str,
		  $f1->smb_remotedir);
	if ($ret != 0) {
		return 1;
	}

	# check top level container 'oi_dir' no longer has OI|CI/READ
	$dir_ace = ace_parse_str($dir_acl_str);
	$ret = file_ace_check($f1->smb_remotedir, $dir_ace);

	if ($ret == 0) {
		say "unexpected ace";
		return 1;
	}

	$child_file_ace = ace_parse_str($file_inherited_ace_str);
	# nested file 'oi_dir/file-1' should NOT have inherited I/READ
	$ret = file_ace_check($f1->smb_remotepath, $child_file_ace);
	if ($ret == 0) {
		say "unexpected ace";
		return 1;
	}

	# nested dir  'oi_dir/nested/' should no longer have OI|CI|I|READ
	$child_dir_ace = ace_parse_str($dir_inherited_ace_str);
	$ret = file_ace_check($f2->smb_remotedir, $child_dir_ace);
	if ($ret == 0) {
		return 1;
	}

	$f1->del_remote();
	$f2->del_remote(1);

	return 0;
}

# test smbcacls '--propagate-inheritance --modify' which attempts to modify the
# ACLfor the file and additionally use inheritance rules to propagate
# appropriate changes to children
#
# This test first adds an ACL with (CI)(OI)(R), then it modifies that acl to be
# (CI)(OI)(D) - where D == 0x00110000
#
# before:
#
#  +-tar_test_dir/    (01)(CI)(I)(F)
#    +-oi_dir/        (CI)(OI)(R)
#    | +-file.1       (I)(R)
#    | +-nested/      (CI)(OI)(I)(R)
#    |   +-file.2     (I)(R)
#
# after/expected:
#
#  +-tar_test_dir/    (01)(CI)(I)(F)
#    +-oi_dir/        (CI)(OI)(CHANGE)
#    | +-file.1       (I)(CHANGE)
#    | +-nested/      (CI)(OI)(I)(CHANGE)
#    |   +-file.2     (I)(CHANGE)

sub test_simple_cioi_modify
{
	my @files;
	my $dir_acl_str = "ACL:$USER:ALLOWED/OI|CI/R";
	my $file_inherited_ace_str = "ACL:$USER:ALLOWED/I/R";
	my $dir_inherited_ace_str = "ACL:$USER:ALLOWED/OI|CI|I/R";

	my $dir_mod_acl_str = "ACL:$USER:ALLOWED/OI|CI/CHANGE";
	my $file_mod_inherited_ace_str = "ACL:$USER:ALLOWED/I/CHANGE";
	my $dir_mod_inherited_ace_str = "ACL:$USER:ALLOWED/OI|CI|I/CHANGE";

	my $dir_ace;
	my $child_file_ace;
	my $child_dir_ace;
	my $ret;
	my $out;

	my $f1 = File->new_local("$TMP/file-1");
	$f1->put_remote("oi_dir/file-1");
	my $f2 = File->new_local("$TMP/file-2");
	$f2->put_remote("oi_dir/nested/file-2");

	# add flags on oi_dir

	# This is somwhat artificial, we need to add a new acl to the directory
	# so that the following modify operation doesn't fail. Previously
	# '--modify' was used in place of '--add' but that resulted in failure
	# to access the directory ( or even modify the acl ).
	# Note: when running this test against a windows server it seems that
	# running as Administrator ensures best results

	($out, $ret) = smb_cacls('--add', $dir_acl_str, $f1->smb_remotedir);
	if ($ret != 0) {
		return 1;
	}
	# add flags on oi_dir/nested
	($out, $ret) = smb_cacls('--add', $dir_inherited_ace_str,
				 $f2->smb_remotedir);
	if ($ret != 0) {
		return 1;
	}
	# add flags on oi_dir/file-1
	($out, $ret) = smb_cacls('--add', $file_inherited_ace_str,
				 $f1->smb_remotepath);
	if ($ret != 0) {
		return 1;
	}
	# add flags on oi_dir/nested/file-2
	($out, $ret) = smb_cacls('--add', $file_inherited_ace_str,
				 $f2->smb_remotepath);
	if ($ret != 0) {
		return 1;
	}
	
	($out, $ret) = smb_cacls('--propagate-inheritance', '--modify',
				 $dir_mod_acl_str,
		  $f1->smb_remotedir);
	if ($ret != 0) {
		return 1;
	}

	# check top level container 'oi_dir' has OI|CI/CHANGE
	$dir_ace = ace_parse_str($dir_mod_acl_str);
	$ret = file_ace_check($f1->smb_remotedir, $dir_ace);
	if ($ret != 0) {
		return 1;
	}

	$child_file_ace = ace_parse_str($file_mod_inherited_ace_str);
	# nested file 'oi_dir/file-1' should have inherited I|CHANGE
	$ret = file_ace_check($f1->smb_remotepath, $child_file_ace);
	if ($ret != 0) {
		say "missing expected ace";
		return 1;
	}

	# nested dir  'oi_dir/nested/' should have OI|CI|I|CHANGE
	$child_dir_ace = ace_parse_str($dir_mod_inherited_ace_str);
	$ret = file_ace_check($f2->smb_remotedir, $child_dir_ace);
	if ($ret != 0) {
		return 1;
	}

	$f1->del_remote();
	$f2->del_remote(1);

	return 0;
}
# test smbcacls '--propagate-inheritance --modify' which attempts to modify ACL
# for the file and additionally use inheritance rules to propagate appropriate
# changes to children
#
# This test first adds an ACL with (OI)(R), then it modifies that acl to be
# (OI)(D) - where D == 0x00110000
#
# before:
#
#  +-tar_test_dir/    (01)(CI)(I)(F)
#    +-oi_dir/        (OI)(IO)(R)
#    | +-file.1       (I)(R)
#    | +-nested/      (OI)(IO)(I)(R)
#    |   +-file.2     (I)(R)
#
# after/expected:
#
#  +-tar_test_dir/    (01)(CI)(I)(F)
#    +-oi_dir/        (OI)(IO)(CHANGE)
#    | +-file.1       (I)(CHANGED)
#    | +-nested/      (OI)(IO)(I)(CHANGED)
#    |   +-file.2     (I)(CHANGED)
sub test_simple_oi_modify
{
	my @files;
	my $explict_access_ace_str = "ACL:$USER:ALLOWED/0x0/RWD";


	my $dir_mod_acl_str = "ACL:$USER:ALLOWED/OI/CHANGE";
	my $file_mod_inherited_ace_str = "ACL:$USER:ALLOWED/I/CHANGE";
	my $dir_mod_inherited_ace_str = "ACL:$USER:ALLOWED/OI|IO|I/CHANGE";

	my $dir_ace;
	my $child_file_ace;
	my $child_dir_ace;
	my $ret;
	my $out;

	my $f1 = File->new_local("$TMP/file-1");
	$f1->put_remote("oi_dir/file-1");
	my $f2 = File->new_local("$TMP/file-2");
	$f2->put_remote("oi_dir/nested/file-2");

	# add flags on oi_dir

	# This is somwhat artificial, we need to add a new acl to the directory
	# so that the following modify operation doesn't fail. Previously
	# '--modify' was used in place of '--add' but that resulted in failure
	# to access the directory ( or even modify the acl ).
	# Note: when running this test against a windows server it seems that
	# running as Administrator ensures best results

	($out, $ret) = smb_cacls('--add', $explict_access_ace_str,
				 $f1->smb_remotedir);
	if ($ret != 0) {
		return 1;
	}
	# add flags on oi_dir/nested
	($out, $ret) = smb_cacls('--add', $explict_access_ace_str,
				 $f2->smb_remotedir);
	if ($ret != 0) {
		return 1;
	}
	# add flags on oi_dir/file-1
	($out, $ret) = smb_cacls('--add', $explict_access_ace_str,
				 $f1->smb_remotepath);
	if ($ret != 0) {
		return 1;
	}
	# add flags on oi_dir/nested/file-2
	($out, $ret) = smb_cacls('--add', $explict_access_ace_str,
				 $f2->smb_remotepath);
	if ($ret != 0) {
		return 1;
	}
	($out, $ret) = smb_cacls('--propagate-inheritance', '--modify',
				 $dir_mod_acl_str, $f1->smb_remotedir);
	if ($ret != 0) {
		return 1;
	}

	# check top level container 'oi_dir' has OI/CHANGE
	$dir_ace = ace_parse_str($dir_mod_acl_str);
	$ret = file_ace_check($f1->smb_remotedir, $dir_ace);
	if ($ret != 0) {
		return 1;
	}
	# file 'oi_dir/file-1' should  have inherited I/CHANGE
	$child_file_ace = ace_parse_str($file_mod_inherited_ace_str);
	$ret = file_ace_check($f1->smb_remotepath, $child_file_ace);

	if ($ret != 0) {
		return 1;
	}

	# nested dir  'oi_dir/nested/' should have OI|IO/CHANGE
	$child_dir_ace = ace_parse_str($dir_mod_inherited_ace_str);
	$ret = file_ace_check($f2->smb_remotedir, $child_dir_ace);
	if ($ret != 0) {
		return 1;
	}

	# nested file 'oi_dir/nested/file-2' should  have inherited I/CHANGE
	$ret = file_ace_check($f2->smb_remotepath, $child_file_ace);
	if ($ret != 0) {
		say "got ace where not expecting";
		return 1;
	}

	$f1->del_remote();
	$f2->del_remote(1);

	return 0;
}
# test smbcacls '--propagate-inheritance --modify' which attempts to modify ACL
# for the file and additionally use inheritance rules to propagate appropriate
# changes to children
#
# This test first adds an ACL with (CI)(R), then it modifies that acl to be
# (CI)(D) - where D == 0x00110000
#
# before:
#
#  +-tar_test_dir/    (01)(CI)(I)(F)
#    +-oi_dir/        (CI)(R)
#    | +-file.1            (I)(F)
#    | +-nested/      (CI)(I)(R)
#    |   +-file.2          (I)(F)
#
# after/expected:
#
#  +-tar_test_dir/    (01)(CI)(I)(F)
#    +-oi_dir/        (CI)(CHANGE)
#    | +-file.1            (I)(F)
#    | +-nested/      (CI)(I)(CHANGE)
#    |   +-file.2          (I)(F)

sub test_simple_ci_modify
{
	my @files;
	my $dir_acl_str = "ACL:$USER:ALLOWED/CI/READ";
	my $file_inherited_ace_str = "ACL:$USER:ALLOWED/I/READ";
	my $dir_inherited_ace_str = "ACL:$USER:ALLOWED/CI|I/READ";

	my $dir_mod_acl_str = "ACL:$USER:ALLOWED/CI/CHANGE";
	my $file_mod_inherited_ace_str = "ACL:$USER:ALLOWED/I/CHANGE";
	my $dir_mod_inherited_ace_str = "ACL:$USER:ALLOWED/CI|I/CHANGE";
	my $delete_ace_str = "ACL:$USER:ALLOWED/0x0/RWD";

	my $dir_ace;
	my $child_file_ace;
	my $child_dir_ace;
	my $ret;
	my $out;

	my $f1 = File->new_local("$TMP/file-1");
	$f1->put_remote("oi_dir/file-1");
	my $f2 = File->new_local("$TMP/file-2");
	$f2->put_remote("oi_dir/nested/file-2");

	# add flags on oi_dir

	# This is somwhat artificial, we need to add a new acl to the directory
	# so that the following modify operation doesn't fail. Previously
	# '--modify' was used in place of '--add' but that resulted in failure
	# to access the directory ( or even modify the acl ).
	# Note: when running this test against a windows server it seems that
	# running as Administrator ensures best results

	($out, $ret) = smb_cacls('--add', $dir_acl_str, $f1->smb_remotedir);
	if ($ret != 0) {
		return 1;
	}
	# add flags on oi_dir/nested
	($out, $ret) = smb_cacls('--add', $dir_inherited_ace_str,
				 $f2->smb_remotedir);
	if ($ret != 0) {
		return 1;
	}

	($out, $ret) = smb_cacls('--propagate-inheritance', '--modify',
				 $dir_mod_acl_str, $f1->smb_remotedir);
	if ($ret != 0) {
		return 1;
	}

	# check top level container 'oi_dir' has CI/CHANGE
	$dir_ace = ace_parse_str($dir_mod_acl_str);
	$ret = file_ace_check($f1->smb_remotedir, $dir_ace);
	if ($ret != 0) {
		return 1;
	}


	$child_file_ace = ace_parse_str($file_mod_inherited_ace_str);
	# nested file 'oi_dir/file-1' should NOT have inherited I/CHANGE
	$ret = file_ace_check($f1->smb_remotepath, $child_file_ace);
	if ($ret == 0) {
		say "got ace where not expecting";
		return 1;
	}

	# nested dir  'oi_dir/nested/' should have OI|I/CHANGE
	$child_dir_ace = ace_parse_str($dir_mod_inherited_ace_str);
	$ret = file_ace_check($f2->smb_remotedir, $child_dir_ace);
	if ($ret != 0) {
		return 1;
	}

	# set some flags to allow us to delete the files
	smb_cacls('--set', $delete_ace_str, $f1->smb_remotepath);
	smb_cacls('--set', $delete_ace_str, $f2->smb_remotepath);
	$f1->del_remote();
	$f2->del_remote(1);

	return 0;
}

# test smbcacls '--propagate-inheritance --set' which attempts to set the ACL
# for the file and additionally use inheritance rules to propagate appropriate
# changes to children
#
# This test adds an ACL with (CI)(OI)(READ)
#
# before:
#
#  +-tar_test_dir/    (01)(CI)(I)(F)
#    +-oi_dir/        (OI)(CI)(I)(F)
#    | +-file.1            (I)(F)
#    | +-nested/      (OI)(CI)(I)(F)
#    |   +-file.2          (I)(F)
#
# after/expected:
# fail, oid_dir has inheritance enabled, set should fail and exit with '1'

sub test_simple_set_fail()
{
	my @files;
	my $dir_acl_str = "ACL:$USER:ALLOWED/OI|CI/R";
	my $file_inherited_ace_str = "ACL:$USER:ALLOWED/I/R";
	my $dir_inherited_ace_str = "ACL:$USER:ALLOWED/OI|CI|I/R";
	my $dir_ace;
	my $child_file_ace;
	my $child_dir_ace;
	my $ret;
	my $out;

	my $f1 = File->new_local("$TMP/file-1");
	$f1->put_remote("oi_dir/file-1");
	my $f2 = File->new_local("$TMP/file-2");
	$f2->put_remote("oi_dir/nested/file-2");

	($out, $ret) = smb_cacls('--propagate-inheritance', '--set',
				 $dir_acl_str, $f1->smb_remotedir);

	if ($ret == 0 ) {
		say "smb_cacls '--set' succeeded unexpectedly while processing container with inheritance enabled";	
		return 1;
	}
	$f1->del_remote();
	$f2->del_remote(1);
	return 0;
}

# test smbcacls '--propagate-inheritance --set' which attempts to set the ACL
# for the file and additionally use inheritance rules to propagate appropriate
# changes to children
#
# This test adds an ACL with (CI)(OI)(RWD) additionally it removes 
# inheritance from oi_dir
#
# before:
#
#  +-tar_test_dir/    (01)(CI)(I)(F)
#    +-oi_dir/        (OI)(CI)(I)(F)
#    | +-file.1            (I)(F)
#    | +-nested/      (OI)(CI)(I)(F)
#    |   +-file.2          (I)(F)
#
# after/expected:
#  +-tar_test_dir/    (01)(CI)(I)(F)
#    +-oi_dir/        (OI)(CI)(RWD)
#    | +-file.1            (I)(RWD)
#    | +-nested/      (OI)(CI)(I)(RWD)
#    |   +-file.2          (I)(RWD)
#

sub test_simple_oici_set()
{
	my @files;
	my $dir_acl_str = "ACL:$USER:ALLOWED/OI|CI/RWD";
	my $file_inherited_ace_str = "ACL:$USER:ALLOWED/I/RWD";
	my $dir_inherited_ace_str = "ACL:$USER:ALLOWED/OI|CI|I/RWD";
	my $dir_ace;
	my $child_file_ace;
	my $child_dir_ace;
	my $ret;
	my $out;

	my $f1 = File->new_local("$TMP/file-1");
	$f1->put_remote("oi_dir/file-1");
	my $f2 = File->new_local("$TMP/file-2");
	$f2->put_remote("oi_dir/nested/file-2");

	($out, $ret) = smb_cacls('--inherit=copy', $f1->smb_remotedir);

	# smb_cacls --inherit=remove 
	if ($ret != 0 ) {
		say "failed to remove inheritance enabled";	
		return 1;
	}

	($out, $ret) = smb_cacls('--propagate-inheritance', '--set',
				 $dir_acl_str, $f1->smb_remotedir);
	if ($ret != 0 ) {
		return 1;
	}
	# check top level container 'oi_dir' has OI|CI/RWD
	$dir_ace = ace_parse_str($dir_acl_str);	
	$ret = file_ace_check($f1->smb_remotedir, $dir_ace);
	if ($ret != 0) {
		return 1;
	}
	# check nested file has I/RW
	$child_file_ace = ace_parse_str($file_inherited_ace_str);	
	$ret = file_ace_check($f1->smb_remotepath, $child_file_ace);
	if ($ret != 0) {
		return 1;
	}

	# check nested dir has OI|CI|I/RW
	$child_dir_ace = ace_parse_str($dir_inherited_ace_str);	
	$ret = file_ace_check($f2->smb_remotedir, $child_dir_ace);
	if ($ret != 0) {
		return 1;
	}
	$f1->del_remote();
	$f2->del_remote(1);
	return 0;
}

# test smbcacls '--propagate-inheritance --set' which attempts to set the ACL
# for the file and additionally use inheritance rules to propagate appropriate
# changes to children
#
# This test adds an ACL with (CI)(RWD) additionally it removes 
# inheritance from oi_dir
#
# before:
#
#  +-tar_test_dir/    (01)(CI)(I)(F)
#    +-oi_dir/        (OI)(CI)(I)(F)
#    | +-file.1            (I)(F)
#    | +-nested/      (OI)(CI)(I)(F)
#    |   +-file.2          (I)(F)
#
# after/expected:
#  +-tar_test_dir/    (01)(CI)(I)(RWD)
#    +-oi_dir/        (CI)(RWD)
#    | +-file.1       
#    | +-nested/      (CI)(I)(RWD)
#    |   +-file.2     
#

sub test_simple_ci_set()
{
	my @files;
	my $dir_acl_str = "ACL:$USER:ALLOWED/CI/RWD";
	my $file_inherited_ace_str = "ACL:$USER:ALLOWED/I/RWD";
	my $dir_inherited_ace_str = "ACL:$USER:ALLOWED/CI|I/RWD";
	my $delete_ace_str = "ACL:$USER:ALLOWED/0x0/RWD";
	my $dir_ace;
	my $child_file_ace;
	my $child_dir_ace;
	my $ret;
	my $out;

	my $f1 = File->new_local("$TMP/file-1");
	$f1->put_remote("oi_dir/file-1");
	my $f2 = File->new_local("$TMP/file-2");
	$f2->put_remote("oi_dir/nested/file-2");

	($out, $ret) = smb_cacls('--inherit=copy', $f1->smb_remotedir);

	# smb_cacls --inherit=remove 
	if ($ret != 0 ) {
		say "failed to remove inheritance enabled";	
		return 1;
	}
	($out, $ret) = smb_cacls('--propagate-inheritance', '--set',
				 $dir_acl_str, $f1->smb_remotedir);
	if ($ret != 0) {
		return 1;
	}

	my $nacl;
	$nacl = num_acls_for_file($f1->smb_remotedir);

	# Although there maybe a couple of users with associated acl(s)
	# before set, after set there should only be 1 acl
	if ($nacl != 1) {
		say "unexpected number of acls ($nacl) after set";
		return 1;
	}

	# check top level container 'oi_dir' has OI|CI/RWD
	$dir_ace = ace_parse_str($dir_acl_str);	
	$ret = file_ace_check($f1->smb_remotedir, $dir_ace);
	if ($ret != 0) {
		return 1;
	}

	# note can't check file because it has no ACL ( due to CI )

	# check nested dir has OI|CI|I/RWD
	$child_dir_ace = ace_parse_str($dir_inherited_ace_str);	
	$ret = file_ace_check($f2->smb_remotedir, $child_dir_ace);
	if ($ret != 0) {
		return 1;
	}
	# set some flags to allow us to delete the files
	smb_cacls('--set', $delete_ace_str, $f1->smb_remotepath);
	smb_cacls('--set', $delete_ace_str, $f2->smb_remotepath);
	$f1->del_remote();
	$f2->del_remote(1);
	return 0;
}

# test smbcacls '--propagate-inheritance --add' which attempts to add the ACL
# for the file and additionally use inheritance rules to propagate appropriate
# changes to children
#
# This test adds an ACL with (CI)(OI)(NP)(CHANGE)
# (NP) - no propagation should not propagate the changes any further containers
#
# before:
#
#  +-tar_test_dir/    (01)(CI)(I)(F)
#    +-oi_dir/        (OI)(CI)(I)(F)
#    | +-file.1            (I)(F)
#    | +-nested/      (OI)(CI)(I)(F)
#    |   +-file.2          (I)(F)
#
# after/expected:
#
#  +-tar_test_dir/    (01)(CI)(I)(F)
#    +-oi_dir/        (OI)(CI)(I)(F), (CI)(OI)(NP)(CHANGE)
#    | +-file.1            (I)(F), (I)(CHANGE)
#    | +-nested/      (OI)(CI)(I)(F), (I)(CHANGE)
#    |   +-file.2          (I)(F)
sub test_simple_cioinp_add
{
	my @files;
	my $dir_add_acl_str = "ACL:$USER:ALLOWED/OI|CI|NP/CHANGE";
	my $inherited_ace_str = "ACL:$USER:ALLOWED/I/CHANGE";
	my $dir_ace;
	my $child_file_ace;
	my $child_dir_ace;
	my $ret;
	my $out;

	my $f1 = File->new_local("$TMP/file-1");
	$f1->put_remote("oi_dir/file-1");
	my $f2 = File->new_local("$TMP/file-2");
	$f2->put_remote("oi_dir/nested/file-2");

	($out, $ret) = smb_cacls('--propagate-inheritance', '--add',
				 $dir_add_acl_str, $f1->smb_remotedir);
	if ($ret != 0) {
		return 1;
	}

	# check top level container 'oi_dir' has OI|CI|NP/READ
	$dir_ace = ace_parse_str($dir_add_acl_str);
	$ret = file_ace_check($f1->smb_remotedir, $dir_ace);
	if ($ret != 0) {
		return 1;
	}


	$child_file_ace = ace_parse_str($inherited_ace_str);
	# nested file 'oi_dir/file-1' should have inherited I/CHANGE
	$ret = file_ace_check($f1->smb_remotepath, $child_file_ace);
	if ($ret != 0) {
		say "missing expected ace";
		return 1;
	}

	# nested dir  'oi_dir/nested/file-2' should NOT have I/CHANGE
	$child_dir_ace = ace_parse_str($inherited_ace_str);
	$ret = file_ace_check($f2->smb_remotepath, $child_dir_ace);
	if ($ret == 0) {
		say "unexpected ace";
		return 1;
	}

	$child_file_ace = ace_parse_str($inherited_ace_str);
	# nested dir 'oi_dir/nested' should have inherited I/CHANGE
	$ret = file_ace_check($f2->smb_remotedir, $child_file_ace);
	if ($ret != 0) {
		say "missing expected ace";
		return 1;
	}

	$f1->del_remote();
	$f2->del_remote(1);

	return 0;
}

# test smbcacls '--propagate-inheritance --add' which attempts to add the ACL
# for the file and additionally use inheritance rules to propagate appropriate
# changes to children
#
# This test adds an ACL with (OI)(NP)(CHANGE)
# (NP) - no propagation should not propagate the changes any further containers
#
# before:
#
#  +-tar_test_dir/    (01)(CI)(I)(F)
#    +-oi_dir/        (OI)(CI)(I)(F)
#    | +-file.1            (I)(F)
#    | +-nested/      (OI)(CI)(I)(F)
#    |   +-file.2          (I)(F)
#
# after/expected:
#
#  +-tar_test_dir/    (01)(CI)(I)(F)
#    +-oi_dir/        (OI)(CI)(I)(F), (OI)(NP)(CHANGE)
#    | +-file.1            (I)(F), (I)(CHANGE)
#    | +-nested/      (OI)(CI)(I)(F)
#    |   +-file.2          (I)(F)
sub test_simple_oinp_add
{
	my @files;
	my $dir_add_acl_str = "ACL:$USER:ALLOWED/OI|NP/CHANGE";
	my $inherited_ace_str = "ACL:$USER:ALLOWED/I/CHANGE";
	my $dir_ace;
	my $child_file_ace;
	my $child_dir_ace;
	my $ret;
	my $out;

	my $f1 = File->new_local("$TMP/file-1");
	$f1->put_remote("oi_dir/file-1");
	my $f2 = File->new_local("$TMP/file-2");
	$f2->put_remote("oi_dir/nested/file-2");

	($out, $ret) = smb_cacls('--propagate-inheritance', '--add',
				 $dir_add_acl_str, $f1->smb_remotedir);
	if ($ret != 0) {
		return 1;
	}

	# check top level container 'oi_dir' has OI|NP/READ
	$dir_ace = ace_parse_str($dir_add_acl_str);
	$ret = file_ace_check($f1->smb_remotedir, $dir_ace);
	if ($ret != 0) {
		return 1;
	}


	$child_file_ace = ace_parse_str($inherited_ace_str);
	# nested file 'oi_dir/file-1' should have inherited I/CHANGE
	$ret = file_ace_check($f1->smb_remotepath, $child_file_ace);
	if ($ret != 0) {
		say "missing expected ace";
		return 1;
	}

	# nested dir  'oi_dir/nested' should NOT have I/CHANGE
	$child_dir_ace = ace_parse_str($inherited_ace_str);
	$ret = file_ace_check($f2->smb_remotedir, $child_dir_ace);
	if ($ret != 0) {
		say "unexpected ace";
		return 1;
	}

	$child_file_ace = ace_parse_str($inherited_ace_str);
	# nested file 'oi_dir/nested/file-1' should NOT have inherited I/CHANGE
	$ret = file_ace_check($f1->smb_remotepath, $child_file_ace);
	if ($ret != 0) {
		say "unexpected ace";
		return 1;
	}

	$f1->del_remote();
	$f2->del_remote(1);

	return 0;
}

# test smbcacls '--propagate-inheritance --add' which attempts to add the ACL
# for the file and additionally use inheritance rules to propagate appropriate
# changes to children
#
# This test adds an ACL with (CI)(NP)(CHANGE)
# (NP) - no propagation should not propagate the changes any further containers
#
# before:
#
#  +-tar_test_dir/    (01)(CI)(I)(F)
#    +-oi_dir/        (OI)(CI)(I)(F)
#    | +-file.1            (I)(F)
#    | +-nested/      (OI)(CI)(I)(F)
#    |   +-file.2          (I)(F)
#
# after/expected:
#
#  +-tar_test_dir/    (01)(CI)(I)(F)
#    +-oi_dir/        (OI)(CI)(I)(F), (CI)(NP)(CHANGE)
#    | +-file.1            (I)(F)
#    | +-nested/      (OI)(CI)(I)(F), (I)(CHANGE)
#    |   +-file.2          (I)(F)
sub test_simple_cinp_add
{
	my @files;
	my $dir_add_acl_str = "ACL:$USER:ALLOWED/CI|NP/CHANGE";
	my $inherited_ace_str = "ACL:$USER:ALLOWED/I/CHANGE";
	my $dir_ace;
	my $child_file_ace;
	my $child_dir_ace;
	my $ret;
	my $out;

	my $f1 = File->new_local("$TMP/file-1");
	$f1->put_remote("oi_dir/file-1");
	my $f2 = File->new_local("$TMP/file-2");
	$f2->put_remote("oi_dir/nested/file-2");

	($out, $ret) = smb_cacls('--propagate-inheritance', '--add',
				 $dir_add_acl_str, $f1->smb_remotedir);
	if ($ret != 0) {
		return 1;
	}

	# check top level container 'oi_dir' has CI|NP/READ
	$dir_ace = ace_parse_str($dir_add_acl_str);
	$ret = file_ace_check($f1->smb_remotedir, $dir_ace);
	if ($ret != 0) {
		return 1;
	}


	$child_file_ace = ace_parse_str($inherited_ace_str);
	# nested file 'oi_dir/file-1' should NOT have inherited I/CHANGE
	$ret = file_ace_check($f1->smb_remotepath, $child_file_ace);
	if ($ret == 0) {
		say "unexpected ace";
		return 1;
	}

	# nested dir  'oi_dir/nested' should have I/CHANGE
	$child_dir_ace = ace_parse_str($inherited_ace_str);
	$ret = file_ace_check($f2->smb_remotedir, $child_dir_ace);
	if ($ret != 0) {
		say "missing expected ace";
		return 1;
	}

	$child_file_ace = ace_parse_str($inherited_ace_str);
	# nested file 'oi_dir/nested/file-1' should NOT have inherited I/CHANGE
	$ret = file_ace_check($f1->smb_remotepath, $child_file_ace);
	if ($ret == 0) {
		say "unexpected ace";
		return 1;
	}

	$f1->del_remote();
	$f2->del_remote(1);
	return 0;
}

# test smbcacls '--propagate-inheritance --delete' which attempts to delete
# the ACL for the file and additionally use inheritance rules to propagate
# appropriate changes to children
#
# This test adds an ACL with (CI)(OI)(NP)(CHANGE)
# (NP) - no propagation should not propagate the changes any further containers
#
# before:
#
#  +-tar_test_dir/    (01)(CI)(I)(F)
#    +-oi_dir/        (OI)(CI)(I)(F), (CI)(OI)(NP)(CHANGE)
#    | +-file.1            (I)(F), (I)(CHANGE)
#    | +-nested/      (OI)(CI)(I)(F), (I)(CHANGE)
#    |   +-file.2          (I)(F)
#
# after/expected:
#
#  +-tar_test_dir/    (01)(CI)(I)(F)
#    +-oi_dir/        (OI)(CI)(I)(F)
#    | +-file.1            (I)(F)
#    | +-nested/      (OI)(CI)(I)(F)
#    |   +-file.2          (I)(F)
#
sub test_simple_cioinp_delete
{
	my @files;
	my $dir_add_acl_str = "ACL:$USER:ALLOWED/OI|CI|NP/CHANGE";
	my $inherited_ace_str = "ACL:$USER:ALLOWED/I/CHANGE";
	my $dir_ace;
	my $child_file_ace;
	my $child_dir_ace;
	my $ret;
	my $out;

	my $f1 = File->new_local("$TMP/file-1");
	$f1->put_remote("oi_dir/file-1");
	my $f2 = File->new_local("$TMP/file-2");
	$f2->put_remote("oi_dir/nested/file-2");

	# set up 'before' permissions
	($out, $ret) = smb_cacls('--add', $dir_add_acl_str, $f1->smb_remotedir);
	if ($ret != 0) {
		return 1;
	}
	($out, $ret) = smb_cacls('--add', $inherited_ace_str,
				 $f1->smb_remotepath);
	if ($ret != 0) {
		return 1;
	}
	($out, $ret) = smb_cacls('--add', $inherited_ace_str,
				 $f2->smb_remotedir);
	if ($ret != 0) {
		return 1;
	}

	($out, $ret) = smb_cacls('--propagate-inheritance', '--delete',
				 $dir_add_acl_str,
		  $f1->smb_remotedir);
	if ($ret != 0) {
		return 1;
	}

	# check top level container 'oi_dir' does NOT have OI|CI|NP/READ
	$dir_ace = ace_parse_str($dir_add_acl_str);
	$ret = file_ace_check($f1->smb_remotedir, $dir_ace);
	if ($ret == 0) {
		say "unexpected ace";
		return 1;
	}


	$child_file_ace = ace_parse_str($inherited_ace_str);
	# nested file 'oi_dir/file-1' should NOT have inherited I/CHANGE
	$ret = file_ace_check($f1->smb_remotepath, $child_file_ace);
	if ($ret == 0) {
		say "unexpected ace";
		return 1;
	}

	$child_file_ace = ace_parse_str($inherited_ace_str);
	# nested dir 'oi_dir/nested' should NOT have inherited I/CHANGE
	$ret = file_ace_check($f2->smb_remotedir, $child_file_ace);
	if ($ret == 0) {
		say "unexpected ace";
		return 1;
	}

	$f1->del_remote();
	$f2->del_remote(1);

	return 0;
}

# test smbcacls '--propagate-inheritance --delete' which attempts to delete the
# ACL for the file and additionally use inheritance rules to propagate
# appropriate changes to children
#
# This test adds an ACL with (CI)(NP)(CHANGE)
# (NP) - no propagation should not propagate the changes any further containers
#
# before:
#
#  +-tar_test_dir/    (01)(CI)(I)(F)
#    +-oi_dir/        (OI)(CI)(I)(F), (CI)(NP)(CHANGE)
#    | +-file.1            (I)(F)
#    | +-nested/      (OI)(CI)(I)(F), (I)(CHANGE)
#    |   +-file.2          (I)(F)
#
# after/expected:
#
#  +-tar_test_dir/    (01)(CI)(I)(F)
#    +-oi_dir/        (OI)(CI)(I)(F)
#    | +-file.1            (I)(F)
#    | +-nested/      (OI)(CI)(I)(F)
#    |   +-file.2          (I)(F)

sub test_simple_cinp_delete
{
	my @files;
	my $dir_add_acl_str = "ACL:$USER:ALLOWED/CI|NP/CHANGE";
	my $inherited_ace_str = "ACL:$USER:ALLOWED/I/CHANGE";
	my $dir_ace;
	my $child_file_ace;
	my $child_dir_ace;
	my $ret;
	my $out;

	my $f1 = File->new_local("$TMP/file-1");
	$f1->put_remote("oi_dir/file-1");
	my $f2 = File->new_local("$TMP/file-2");
	$f2->put_remote("oi_dir/nested/file-2");

	# set up ace(s) to match 'before'
	($out, $ret) = smb_cacls('--add', $dir_add_acl_str, $f1->smb_remotedir);
	if ($ret != 0) {
		return 1;
	}
	($out, $ret) = smb_cacls('--add', $inherited_ace_str,
				 $f2->smb_remotedir);
	if ($ret != 0) {
		return 1;
	}

	($out, $ret) = smb_cacls('--propagate-inheritance', '--delete',
				 $dir_add_acl_str, $f1->smb_remotedir);
	if ($ret != 0) {
		return 1;
	}

	# check top level container 'oi_dir' doesn't have CI|NP/READ
	$dir_ace = ace_parse_str($dir_add_acl_str);
	$ret = file_ace_check($f1->smb_remotedir, $dir_ace);
	if ($ret == 0) {
		say "unexpected ace";
		return 1;
	}


	$child_file_ace = ace_parse_str($inherited_ace_str);
	# nested file 'oi_dir/file-1' should NOT have inherited I/CHANGE
	$ret = file_ace_check($f1->smb_remotepath, $child_file_ace);
	if ($ret == 0) {
		say "unexpected ace";
		return 1;
	}

	# nested dir  'oi_dir/nested' should NOT have I/CHANGE
	$child_dir_ace = ace_parse_str($inherited_ace_str);
	$ret = file_ace_check($f2->smb_remotedir, $child_dir_ace);
	if ($ret == 0) {
		say "unexpected ace";
		return 1;
	}

	$child_file_ace = ace_parse_str($inherited_ace_str);
	# nested file 'oi_dir/nested/file-1' should NOT have inherited I/CHANGE
	$ret = file_ace_check($f1->smb_remotepath, $child_file_ace);
	if ($ret == 0) {
		say "unexpected ace";
		return 1;
	}

	$f1->del_remote();
	$f2->del_remote(1);

	return 0;
}

# test smbcacls '--propagate-inheritance --delete' which attempts to delete the
# ACL for the file and additionally use inheritance rules to propagate
# appropriate changes to children
#
# This test adds an ACL with (OI)(NP)(CHANGE)
# (NP) - no propagation should not propagate the changes any further containers
#
# before:
#
#  +-tar_test_dir/    (01)(CI)(I)(F)
#    +-oi_dir/        (OI)(CI)(I)(F), (OI)(NP)(CHANGE)
#    | +-file.1            (I)(F), (I)(CHANGE)
#    | +-nested/      (OI)(CI)(I)(F)
#    |   +-file.2          (I)(F)
#
# after/expected:
#
#  +-tar_test_dir/    (01)(CI)(I)(F)
#    +-oi_dir/        (OI)(CI)(I)(F)
#    | +-file.1            (I)(F)
#    | +-nested/      (OI)(CI)(I)(F)
#    |   +-file.2          (I)(F)

sub test_simple_oinp_delete
{
	my @files;
	my $dir_add_acl_str = "ACL:$USER:ALLOWED/OI|NP/CHANGE";
	my $inherited_ace_str = "ACL:$USER:ALLOWED/I/CHANGE";
	my $dir_ace;
	my $child_file_ace;
	my $child_dir_ace;
	my $ret;
	my $out;

	my $f1 = File->new_local("$TMP/file-1");
	$f1->put_remote("oi_dir/file-1");
	my $f2 = File->new_local("$TMP/file-2");
	$f2->put_remote("oi_dir/nested/file-2");

	# set up 'before' permissions
	($out, $ret) = smb_cacls('--add', $dir_add_acl_str, $f1->smb_remotedir);
	if ($ret != 0) {
		return 1;
	}
	($out, $ret) = smb_cacls('--add', $inherited_ace_str,
				 $f1->smb_remotepath);
	if ($ret != 0) {
		return 1;
	}

	($out, $ret) = smb_cacls('--propagate-inheritance', '--delete',
				 $dir_add_acl_str, $f1->smb_remotedir);
	if ($ret != 0) {
		return 1;
	}

	# check top level container 'oi_dir' does NOT have OI|NP/READ
	$dir_ace = ace_parse_str($dir_add_acl_str);
	$ret = file_ace_check($f1->smb_remotedir, $dir_ace);
	if ($ret == 0) {
		say "unexpected ace";
		return 1;
	}


	$child_file_ace = ace_parse_str($inherited_ace_str);
	# nested file 'oi_dir/file-1' should NOT have inherited I/CHANGE
	$ret = file_ace_check($f1->smb_remotepath, $child_file_ace);
	if ($ret == 0) {
		say "unexpected ace";
		return 1;
	}

	$f1->del_remote();
	$f2->del_remote(1);

	return 0;
}

# test smbcacls '--propagate-inheritance --add' which attempts to add the ACL
# for the file and additionally use inheritance rules to propagate appropriate
# changes to children. In particular it tests that inheritance removed does
# indeed prevent inheritance propagation
#
# This test adds an ACL with (CI)(OI)(CHANGE) at oi_dir
#
# Note: Inheritance has been removed ( and ace(s) copied ) at
#   tar_test_dir/oi_dir/nested
#
# before:
#
#  +-tar_test_dir/    (01)(CI)(I)(F)
#    +-oi_dir/        (OI)(CI)(I)(F)
#    | +-file.1            (I)(F)
#    | +-nested/      (OI)(CI)(F)
#    |   +-file.2          (I)(F)
#
# after/expected:
#
#  +-tar_test_dir/    (01)(CI)(I)(F)
#    +-oi_dir/        (OI)(CI)(I)(F), (CI)(OI)(CHANGE)
#    | +-file.1            (I)(F), (I)((CHANGE)
#    | +-nested/      (OI)(CI)(F)
#    |   +-file.2          (I)(F)
sub test_simple_cioi_inhibit
{
	my @files;
	my $dir_add_acl_str = "ACL:$USER:ALLOWED/OI|CI/CHANGE";
	my $file_inherited_ace_str = "ACL:$USER:ALLOWED/I/CHANGE";
	my $dir_inherited_ace_str = "ACL:$USER:ALLOWED/OI|CI|I/CHANGE";
	my $dir_ace;
	my $child_file_ace;
	my $child_dir_ace;
	my $ret;
	my $out;

	my $f1 = File->new_local("$TMP/file-1");
	$f1->put_remote("oi_dir/file-1");
	my $f2 = File->new_local("$TMP/file-2");
	$f2->put_remote("oi_dir/nested/file-2");

	($out, $ret) = smb_cacls('--inherit=copy', $f2->smb_remotedir);

	# smb_cacls --inherit=remove 
	if ($ret != 0 ) {
		say "failed to remove inheritance enabled";	
		return 1;
	}

	smb_cacls('--propagate-inheritance', '--add', $dir_add_acl_str,
		  $f1->smb_remotedir);

	# check top level container 'oi_dir' has OI|CI/CHANGE
	$dir_ace = ace_parse_str($dir_add_acl_str);
	$ret = file_ace_check($f1->smb_remotedir, $dir_ace);
	if ($ret != 0) {
		say "missing expected ace";
		return 1;
	}


	$child_file_ace = ace_parse_str($file_inherited_ace_str);
	# nested file 'oi_dir/file-1' should have inherited I/CHANGE
	$ret = file_ace_check($f1->smb_remotepath, $child_file_ace);
	if ($ret != 0) {
		say "missing expected ace";
		return 1;
	}

	# nested dir  'oi_dir/nested/' should NOT have OI|CI|I/CHANGE
	$child_dir_ace = ace_parse_str($dir_inherited_ace_str);
	$ret = file_ace_check($f2->smb_remotedir, $child_dir_ace);
	if ($ret == 0) {
		say "unexpected ace";
		return 1;
	}

	# nested file  'oi_dir/nested/file-2' should NOT have I/CHANGE
	$child_dir_ace = ace_parse_str($dir_inherited_ace_str);
	$ret = file_ace_check($f2->smb_remotepath, $child_dir_ace);
	if ($ret == 0) {
		say "unexpected ace";
		return 1;
	}
	$f1->del_remote();
	$f2->del_remote(1);

	return 0;
}

#################################

# IMPLEMENTATION

=head2 Useful functions

Here are a list of useful functions and helpers to define tests.

=cut

# list test number and description
sub list_test {
    my $i = 0;
    for (@TESTS) {
        my ($desc, $f, @args) = @$_;
        printf "%2d.\t%s\n", $i++, $desc;
    }
}

sub run_test {
    if ($SUBUNIT) {
        run_test_subunit(@_);
    } else {
        run_test_normal(@_);
    }
}

sub run_test_normal {
    for (@_) {
        my ($desc, $f, @args) = @$_;
        my $err;

        reset_env();
        say "TEST: $desc";
        if ($VERBOSE) {
            $err = $f->(@args);
        } else {
            # turn off STDOUT
            open my $saveout, ">&STDOUT";
            open STDOUT, '>', File::Spec->devnull();
            $err = $f->(@args);
            open STDOUT, ">&", $saveout;
        }
        print_res($err);
        print "\n";
    }
    reset_env();
}

sub run_test_subunit {
    for (@_) {
        my ($desc, $f, @args) = @$_;
        my $err;
        my $str = '';

        reset_env();
        say "test: $desc";

        # capture output in $buf
        my $buf = '';
        open my $handle, '>', \$buf;
        select $handle;

        # check for die() calls
        eval {
            $err = $f->(@args);
        };
        if ($@) {
            $str = $@;
            $err = 1;
        }
        close $handle;

        # restore output
        select STDOUT;

        # result string is output + eventual exception message
        $str = $buf.$str;

        printf "%s: %s [\n%s]\n", ($err > 0 ? "failure" : "success"), $desc, $str;
    }
    reset_env();
}

sub parse_test_string {
    my $s = shift;
    my @tests = ();

    if (!length($s)) {
        return ();
    }

    for (split /,/, $s) {
        if (/^\d+$/) {
            if ($_ >= @TESTS) {
                return ();
            }
            push @tests, $TESTS[$_];
        }
        elsif (/^(\d+)-(\d+)$/) {
            my ($min, $max) = sort ($1, $2);
            if ($max >= @TESTS) {
                return ();
            }

            for ($min..$max) {
                push @tests, $TESTS[$_];
            }
        }
        else {
            return ();
        }
    }

    return @tests;
}

sub print_res {
    my $err = shift;
    if ($err) {
        printf " RES: %s%d ERR%s\n", color('bold red'), $err, color 'reset';
    } else {
        printf " RES: %sOK%s\n", color('bold green'), color 'reset';
    }
}

sub make_env {
    my ($exts, $dirs) = @_;
    my @all;
    my $nb = 0;
    for my $dir (@$dirs) {
        for (@$exts) {
            my $fn = $dir . "file$nb." . $_;
            my $f = File->new_remote($fn, 'ABSPATH');
            $f->delete_on_destruction(1);
            push @all, $f;
            $nb++;
        }
    }

    @all;
}

=head3 C<combine ( \@set, $n )>

=head3 C<combine ( ['a', 'b', 'c'], 2 )>

Return a list of all possible I<n>-uplet (or combination of C<$n> element) of C<@set>.

=cut
sub combine {
    my ($list, $n) = @_;
    die "Insufficient list members" if $n > @$list;

    return map [$_], @$list if $n <= 1;

    my @comb;

    for (my $i = 0; $i+$n <= @$list; $i++) {
        my $val = $list->[$i];
        my @rest = @$list[$i+1..$#$list];
        push @comb, [$val, @$_] for combine(\@rest, $n-1);
    }

    return @comb;
}


=head3 C<reset_remote( )>

Remove all files in the server C<$DIR> (not root)

=cut
sub reset_remote {
    # remove_tree($LOCALPATH . '/'. $DIR);
    # make_path($LOCALPATH . '/'. $DIR);
    if ($CLEAN) {
	    remove_tree($LOCALPATH, {keep_root => 1});
    }
    make_path($LOCALPATH, {keep_root => 1});
}

=head3 C<reset_tmp( )>

Remove all files in the temp directory C<$TMP>

=cut
sub reset_tmp {
    remove_tree($TMP, { keep_root => 1 });
}


=head3 C<reset_env( )>

Remove both temp and remote (C<$DIR>) files

=cut
sub reset_env {
    reset_tmp();
    reset_remote();
}

=head3 C<file_list ( @files )>

Make a multiline string of all the files remote path, one path per line.

C<@files> must be a list of C<File> instance.

=cut
sub file_list {
    my @files = @_;
    my $s = '';
    for (@files) {
        $s .= $_->remotepath."\n";
    }
    return $s;
}

# remove leading "./"
sub remove_dot {
    my $s = shift;
    $s =~ s{^\./}{};
    $s;
}

=head3 C<check_remote( $remotepath, \@files )>

Check if C<$remotepath> has B<exactly> all the C<@files>.

Print a summary on STDOUT.

C<@files> must be a list of C<File> instance.

=cut
sub check_remote {
    my ($subpath, $files) = @_;
    my (%done, %expected);
    my (@less, @more, @diff);

    for (@$files) {
        my $fn = remove_dot($_->remotepath);
        $expected{$fn} = $_;
        $done{$fn} = 0;
    }

    my %remote;
    File::walk(sub { $remote{remove_dot($_->remotepath)} = $_ }, File::tree($subpath));

    for my $rfile (sort keys %remote) {

        # files that shouldn't be there
        if (!exists $expected{$rfile}) {
            say " +    $rfile";
            push @more, $rfile;
            next;
        }

        # same file multiple times
        if ($done{$rfile} > 0) {
            $done{$rfile}++;
            push @more, $rfile;
            printf " +%3d %s\n", $done{$rfile}, $rfile;
            next;
        }

        $done{$rfile}++;

        # different file
        my $rmd5 = $remote{$rfile}->md5;
        if ($expected{$rfile}->md5 ne $rmd5) {
            say " !    $rfile ($rmd5)";
            push @diff, $rfile;
            next;
        }

        say "      $rfile";
    }

    # file that should have been in tar
    @less = grep { $done{$_} == 0 } sort keys %done;
    for (@less) {
        say " -    $_";
    }

    # summary
    printf("\t%d files, +%d, -%d, !%d\n",
           scalar keys %done,
           scalar @more,
           scalar @less,
           scalar @diff);
    return (@more + @less + @diff); # nb of errors
}

sub ace_dump
{
	my ($ace) = @_;
	my $key;
	my $value;

	while (($key, $value) = each(%$ace)) {
		print $key . "=" . $value . ", ";
	}
	print "\n";
}

sub ace_cmp
{
	my ($ace_left, $ace_right) = @_;
	my $key;
	my $value;
	# need to copy hashes, otherwise entries are lost on failure!
	my %ace1 = %$ace_left;
	my %ace2 = %$ace_right;

	while (($key, $value) = each(%ace1)) {
		# ignore the users domain, only compare username
		next if ($key eq "user_dom");

		unless (defined $ace2{$key}) {
			print "no entry for: $key\n";
			return 1;
		}
		if ($value ne $ace2{$key}) {
			print "mismatch: $key:$value != $key:$ace2{$key}\n";
			return 1;
		}
	}

	return 0;
}

=head3 C<ace_parse_str( $ace_str )>

Parses the smbacls C<$ace_str> in the format of:
	ACL:SAMBA-TEST\ddiss:ALLOWED/0x0/FULL
The ACE is returned as hashref with the following members:
->user_dom:	Domain assigned to ACE user, if any
->user:		ACE user, without domain portion
->type:		allow or deny
->inherit:	inheritace flags
->permissions:	permissions flags

=cut
sub ace_parse_str
{
	my ($ace_str) = @_;
	my %ace;

	my @splat = split(":", $ace_str);

	if ($splat[0] ne "ACL") {
		say "invalid ACE string: $ace_str";
		return undef;
	}

	#check for domain component in username
	if (index($splat[1], "\\") >= 0) {
		($ace{user_dom}, $ace{user}) = split(/\\/, $splat[1]);
	} elsif (index($splat[1], "/") >= 0) {
		($ace{user_dom}, $ace{user}) = split(m%/%, $splat[1]);
	} else {
		$ace{user} = $splat[1];
	}

	($ace{type}, $ace{inherit}, $ace{permissions}) = split(/\//, $splat[2]);

	return \%ace;
}
=head3 C<num_acls_for_file( $file )>

count ACL that C<$file> has

C<$file> must be a C<File> instance.

=cut
sub num_acls_for_file
{
	my ($file_remotepath) = @_;
	my @aces;
	my @out;
	my ($output, $ret) =  smb_cacls("--get", $file_remotepath); 
	@out = split("\n", $output);
	@aces = grep(/^ACL/, @out);
	return scalar @aces;
}
=head3 C<file_ace_check( $file, $ace )>

Check that C<$file> has an ACE matching C<$ace>.

C<$file> must be a C<File> instance.

=cut
sub file_ace_check
{
	my ($file_remotepath, $ace) = @_;
	my $missing = 1;
	my @aces;
	my @out;
	my ($output, $ret) =  smb_cacls("--get", $file_remotepath); 
	@out = split("\n", $output);
	@aces = grep(/^ACL/, @out);
	# trim the domain portion of users
	foreach (@aces) {
		my $acl_ace = ace_parse_str($_);
		return 1 unless defined $acl_ace;
		# only compare aces for the same user and type
		next unless (($acl_ace->{user} eq $ace->{user})
			  && ($acl_ace->{type} eq $ace->{type}));

		say "found ACE for $acl_ace->{user}";
		if (ace_cmp($acl_ace, $ace) != 0) {
			print "differences between file ACE: ";
			ace_dump($acl_ace);
			print "and expected ACE: ";
			ace_dump($ace);
		}
		else {
			say "matched ACE for $acl_ace->{user}";
			ace_dump($ace);
			$missing = 0;
			last;
		}
	}

	return $missing;
}

=head3 C<smb_client ( @args )>

Run smbclient with C<@args> passed as argument and return output.

Each element of C<@args> becomes one escaped argument of smbclient.

Host, share, user, password and the additionnal arguments provided on
the command-line are already inserted.

The output contains both the C<STDOUT> and C<STDERR>.

Die if smbclient crashes or exits with an error code.

=cut
sub smb_client {
    my (@args) = @_;

    my $fullpath = "//$HOST/$SHARE";
    my $cmd = sprintf("%s %s %s",
                      quotemeta($CLI_BIN),
                      quotemeta($fullpath),
                      join(' ', map {quotemeta} (@SMBARGS, @args)));

    if ($DEBUG) {
        my $tmp = $cmd;
        $tmp =~ s{\\([./+-])}{$1}g;
        say color('bold yellow'), $tmp, color('reset');
    }

    my $out = `$cmd 2>&1`;
    my $err = $?;
    my $errstr = '';
    # handle abnormal exit
    if ($err == -1) {
        $errstr = "failed to execute $cmd: $!\n";
    }
    elsif ($err & 127) {
        $errstr = sprintf "child died with signal %d (%s)\n", ($err & 127), $cmd;
    }
    elsif ($err >> 8) {
        $errstr = sprintf "child exited with value %d (%s)\n", ($err >> 8), $cmd;
    }

    if ($DEBUG) {
        say $out;
    }

    if ($err) {
        die "ERROR: $errstr";
    }
    return $out;
}

=head3 C<smb_cacls ( @args )>

Run smbcacls with C<@args> passed as argument and return output.

Each element of C<@args> becomes one escaped argument of smbcacls.

Host, share, user, password and the additionnal arguments provided on
the command-line are already inserted.

The output contains both the C<STDOUT> and C<STDERR>.

Die if smbcacls crashes or exits with an error code.

=cut
sub smb_cacls
{
	my (@args) = @_;

	# last arg is the filename, share path must be immediately prior
	my $file = pop(@args);
	my $sharepath = "//$HOST/$SHARE";
	my $cmd = sprintf("%s %s %s %s",
			  quotemeta($CACLS_BIN),
			  join(' ', map {quotemeta} (@SMBARGS, @args)),
			  quotemeta($sharepath),
			  quotemeta($file));

	if ($DEBUG) {
		my $tmp = $cmd;
		$tmp =~ s{\\([./+-])}{$1}g;
		say color('bold yellow'), $tmp, color('reset');
	}

	my $out = `$cmd 2>&1`;
	my $err = $?;
	my $errstr = '';
	my $ret = 0;
	# handle abnormal exit
	if ($err == -1) {
		$errstr = "failed to execute $cmd: $!\n";
	} elsif ($err & 127) {
		$errstr = sprintf "child died with signal %d (%s)\n", ($err & 127), $cmd;
	} elsif ($err >> 8) {
		$errstr = sprintf "child exited with value %d (%s)\n", ($err >> 8), $cmd;
		$ret = ($err >> 8);
	}

	if ($DEBUG) {
		say $out;
	}

	if ($err) {
		# die for abnormal error (cmd didn't run or some signal caught)
		die "ERROR: $errstr" if $ret == 0;
	}
	return ($out, $ret);
}

=head3 C<random( $min, $max )>

Return integer in C<[ $min ; $max ]>

=cut
sub random {
    my ($min, $max) = @_;
    ($min, $max) = ($max, $min) if ($min > $max);
    $min + int(rand($max - $min));
}

#################################

package File;

=head2 The File module

All the test should use the C<File> class. It has nice functions and
methods to deal with paths, to create random files, to list the
content of the server, to change attributes, etc.

There are 2 kinds of C<File>: remote and local.

=over

=item * Remote files are accessible on the server.

=item * Local files are not.

=back

Thus, some methods only works on remote files. If they do not make
sense for local ones, they always return undef.

=cut
use File::Basename;
use File::Path qw/make_path remove_tree/;
use Digest::MD5 qw/md5_hex/;
use Scalar::Util 'blessed';

=head3 Constructors

=head4 C<< File->new_remote($path [, $abs, $size]) >>

Creates a file accessible on the server at C<$DIR/$path> ie. not at the
root of the share and write C<$size> random bytes.

If no size is provided, a random size is chosen.

If you want to remove the automatic prefix C<$DIR>, set C<$abs> to 1.

The file is created without any DOS attributes.

If C<$path> contains non-existent directories, they are automatically
created.

=cut
sub new_remote {
    my ($class, $path, $abs, $size) = @_;
    my ($file, $dir) = fileparse($path);

    $dir = '' if $dir eq './';
    my $loc;

    if ($abs) {
        $loc = cleanpath($main::LOCALPATH.'/'.$dir);
    } else {
        $dir = cleanpath($main::DIR.'/'.$dir);
        $loc = cleanpath($main::LOCALPATH.'/'.$dir);
    }

    make_path($loc);

    my $self = bless {
        'attr' => {qw/r 0 s 0 h 0 a 0 d 0 n 0/},
        'dir'  => $dir,
        'name' => $file,
        'md5'  => create_file($loc.'/'.$file, $size),
        'remote' => 1,
    }, $class;

    $self->set_attr();

    $self;
}

=head4 C<< File->new_local($abs_path [, $data]) >>

Creates a file at C<$abs_path> with $data in it on the system.
If $data is not provided, fill it with random bytes.

=cut
sub new_local {
    my ($class, $path, $data) = @_;
    my ($file, $dir) = fileparse($path);

    make_path($dir);

    my $md5;

    if (defined $data) {
        open my $f, '>', $path or die "can't write in $path: $!";
        print $f $data;
        close $f;
        $md5 = md5_hex($data);
    } else {
        $md5 = create_file($path);
    }

    my $self = {
        'attr' => {qw/r 0 s 0 h 0 a 0 d 0 n 0/},
        'dir'  => $dir,
        'name' => $file,
        'md5'  => $md5,
        'remote' => 0,
    };

    bless $self, $class;
}

=head3 Methods

=head4 C<< $f->localpath >>

Return path on the system eg. F</opt/samba/share/test_tar_mode/file>

=cut
sub localpath {
    my $s = shift;
    if ($s->{remote}) {
        return cleanpath($main::LOCALPATH.'/'.$s->remotepath);
    }
    else {
        return cleanpath($s->{dir}.'/'.$s->{name});
    }
}

=head4 C<< $f->remotepath >>

Return path on the server share.

Return C<undef> if the file is local.

=cut
sub remotepath {
    my ($s) = @_;
    return undef if !$s->{remote};
    my $r = $s->{dir}.'/'.$s->{name};
    $r =~ s{^/}{};
    return cleanpath($r);
}

=head4 C<< $f->smb_remotepath >>

Return SMB path on the server share.

Return C<undef> if the file is local.

=cut
sub smb_remotepath {
    my ($s) = @_;
    return undef if !$s->{remote};
    my $r = $s->remotepath;
    $r =~ s{/}{\\}g;
    return $r;
}

=head4 C<< $f->remotedir >>

Return the directory path of the file on the server.

Like C<< $f->remotepath >> but without the final file name.

=cut
sub remotedir {
    my $s = shift;
    return undef if !$s->{remote};
    cleanpath($s->{dir});
}

=head4 C<< $f->smb_remotepath >>

Return SMB directory on the server share.

Return C<undef> if the file is local.

=cut
sub smb_remotedir {
    my ($s) = @_;
    return undef if !$s->{remote};
    my $r = $s->remotedir;
    $r =~ s{/}{\\}g;
    return $r;
}

=head4 C<< $f->tarpath >>

Return path as it would appear in a tar archive.

Like C<< $f->remotepath >> but prefixed with F<./>

=cut
sub tarpath {
    my $s = shift;
    return undef if !$s->{remote};
    my $r = $s->remotepath;
    $r =~ s{^\./+}{};
    return $r;
}

=head4 C<< $f->delete_on_destruction( 0 ) >>

=head4 C<< $f->delete_on_destruction( 1 ) >>

By default, a C<File> is not deleted on the filesystem when it is not
referenced anymore in Perl memory.

When set to 1, the destructor unlink the file if it is not already removed.
If the C<File> created directories when constructed, it does not remove them.

=cut
sub delete_on_destruction {
    my ($s, $delete) = @_;
    $s->{delete_on_destruction} = $delete;
}

=head4 C<< $f->set_attr( ) >>

=head4 C<< $f->set_attr( 'a' ) >>

=head4 C<< $f->set_attr( 'a', 'r', 's', 'h' ) >>

Remove all DOS attributes and only set the one provided.

=cut
sub set_attr {
    my ($s, @flags) = @_;
    return undef if !$s->{remote};

    $s->{attr} = {qw/r 0 s 0 h 0 a 0 d 0 n 0/};

    for (@flags) {
        $s->{attr}{lc($_)} = 1;
    }

    my $file = $s->{name};
    my @args;
    if ($s->remotedir) {
        push @args, '-D', $s->remotedir;
    }
    main::smb_client(@args, '-c', qq{setmode "$file" -rsha});
    if (@flags && $flags[0] !~ /n/i) {
        main::smb_client(@args, '-c', qq{setmode "$file" +}.join('', @flags));
    }
}

=head4 C<< $f->attr_any( 'a' ) >>

=head4 C<< $f->attr_any( 'a', 's', ... ) >>

Return 1 if the file has any of the DOS attributes provided.

=cut
sub attr_any {
    my ($s, @flags) = @_;
    for (@flags) {
        return 1 if $s->{attr}{$_};
    }
    0;
}


=head4 C<< $f->attr( 'a' ) >>

=head4 C<< $f->attr( 'a', 's', ... ) >>

Return 1 if the file has all the DOS attributes provided.

=cut
sub attr {
    my ($s, @flags) = @_;
    for (@flags) {
        return 0 if !$s->{attr}{$_};
    }
    1;
}

=head4 C<< $f->attr_str >>

Return DOS attributes as a compact string.

  Read-only, hiden, system, archive => "rhsa"

=cut
sub attr_str {
    my $s = shift;
    return undef if !$s->{remote};
    join('', map {$_ if $s->{attr}{$_}} qw/r h s a d n/);
}

=head4 C<< $f->set_time($t) >>

Set modification and access time of the file to C<$t>.

C<$t> must be in Epoch time (number of seconds since 1970/1/1).

=cut
sub set_time {
    my ($s, $t) = @_;
    utime $t, $t, $s->localpath;
}

=head4 C<< $f->md5 >>

Return md5 sum of the file.

The result is cached.

=cut
sub md5 {
    my $s = shift;

    if (!$s->{md5}) {
        open my $h, '<', $s->localpath() or die "can't read ".$s->localpath.": $!";
        binmode $h;
        $s->{md5} = Digest::MD5->new->addfile($h)->hexdigest;
        close $h;
    }

    return $s->{md5};
}

=head4 C<< $f->put_remote($path, [$abs]) >>

Transfers an existing local file to the server at C<$DIR/$path>.
The local file is removed on success.

=cut
sub put_remote
{
	my ($s, $path, $abs) = @_;
	my ($file, $dir) = fileparse($path);
	my $remote_path = "";

	return undef if $s->{remote};

	$dir = '' if $dir eq './';
	my $loc;

	if ($abs) {
		$loc = cleanpath($main::LOCALPATH.'/'.$dir);
	} else {
		$dir = cleanpath($main::DIR.'/'.$dir);
		$loc = cleanpath($main::LOCALPATH.'/'.$dir);
	}

	if ($dir) {
		foreach (split("\/", $dir)) {
			$remote_path .= $_ . "\\";
			main::smb_client('-c', qq{mkdir "$remote_path"});
		}
	}

	$remote_path .= $file;
	my $local_path = $s->localpath();
	main::smb_client('-c', qq{put "$local_path" "$remote_path"});

	$s->{dir} = $dir;
	$s->{name} = $file;
	$s->{remote} = 1;

	return 0;
}

=head4 C<< $f->del_remote() >>

Delete remote file.

=cut
sub del_remote
{
	my ($s, $rm_parent_dirs) = @_;

	return 1 unless $s->{remote};

	my $remote_path = $s->smb_remotepath;

	main::smb_client('-c', qq{rm "$remote_path"});

	unless ($rm_parent_dirs) {
		return 0;
	}

	# attempt to remove all parent dirs
	my @cmps = split(/\\/, $s->smb_remotedir);
	while (scalar(@cmps) > 0) {
		my $path = join("\\", @cmps);
		main::smb_client('-c', qq{rmdir "$path"});
		pop(@cmps);
	}

	return 0;
}

sub DESTROY {
    my $s = shift;
    if ($s->{delete_on_destruction} && -f $s->localpath) {
        if ($main::DEBUG) {
            say "DESTROY ".$s->localpath;
        }
        unlink $s->localpath;
    }
}

=head3 Functions

=head4 C<< File::walk( \&function, @files) >>

=head4 C<< File::walk( sub { ... }, @files) >>

Iterate on file hierachy in C<@files> and return accumulated results.

Use C<$_> in the sub to access the current C<File>.

The C<@files> must come from a call to the C<File::tree> function.

=cut
sub walk {
    my $fun = \&{shift @_};

    my @res;

    for (@_) {
        if ($_->{attr}{d}) {
            push @res, walk($fun, @{$_->{content}});
        } else {
            push @res, $fun->($_);
        }
    }

    return @res;
}

=head4 C<< File::list( $remotepath ) >>

Return list of file (C<File> instance) in C<$remotepath>.

C<$remotepath> must be a directory.

=cut
sub list {
    my ($path) = @_;
    $path ||= '/';
    my @files;
    my $out = main::smb_client('-D', $path, '-c', 'ls');
    $path =~ s{^/}{};

    for (split /\n/, $out) {
        next if !/^  (.+?)\s+([AHSRDN]*)\s+(\d+)\s+(.+)/o;
        my ($fn, $attr, $size, $date) = ($1, $2, $3, $4);
        next if $fn =~ /^\.{1,2}$/;

        push @files, bless {
            'remote' => 1,
            'dir'    => $path,
            'name'   => $fn,
            'size'   => int($size),
            'date'   => $date,
            'attr'   => {
                # list context returns something different than the
                # boolean matching result => force scalar context
                'a' => scalar ($attr =~ /A/),
                'h' => scalar ($attr =~ /H/),
                's' => scalar ($attr =~ /S/),
                'r' => scalar ($attr =~ /R/),
                'd' => scalar ($attr =~ /D/),
                'n' => scalar ($attr =~ /N/),
            },
        }, 'File';
    }
    return @files;
}

=head4 C<< File::tree( $remotepath ) >>

Return recursive list of file in C<$remotepath>.

C<$remotepath> must be a directory.

Use C<File::walk()> to iterate over all the files.

=cut
sub tree {
    my ($d) = @_;
    my @files;

    if (!defined $d) {
        @files = list();
    } elsif (blessed $d) {
        @files = list($d->remotepath);
    } else {
        @files = list($d);
    }

    for my $f (@files) {
        if ($f->{attr}{d}) {
            $f->{content} = [tree($f)];
        }
    }

    return @files;
}

# remove trailing or duplicated slash
sub cleanpath {
    my $p = shift;
    $p =~ s{/+}{/}g;
    $p =~ s{/$}{};
    $p;
}

# create random file at path local path $fn
sub create_file {
    my ($fn, $size) = @_;
    my $buf = '';
    unlink $fn if -e $fn;
    $size ||= main::random(512, 1024);
    $size = int($size);
    my $md5;

    # try /dev/urandom, faster
    if (-e '/dev/urandom') {
        my $cmd = sprintf('head -c %d /dev/urandom | tee %s | md5sum',
                          $size, quotemeta($fn));
        $md5 = (split / /, `$cmd`)[0];
    } else {
        open my $out, '>', $fn or die "can't open $fn: $!\n";
        binmode $out;
        for (1..$size) {
            $buf .= pack('C', main::random(0, 256));
        }
        print $out $buf;
        close $out;
        $md5 = md5_hex($buf);
    }
    return $md5;
}


=head3 Examples

    # create remote file in $DIR/foo/bar
        my $f = File->new_remote("foo/bar/myfile");
        say $f->localpath;  # /opt/share/$DIR/foo/bar/myfile
        say $f->remotepath; # $DIR/foo/bar/myfile
        say $f->remotedir;  # $DIR/foo/bar


    # same but in root dir
        my $f = File->new_remote("myfile", 1);
        say $f->localpath;  # /opt/share/myfile
        say $f->remotepath; # myfile
        say $f->remotedir;  #


    # create local random temp file in $TMP
        my $f = File->new_local("$TMP/temp");
        say $f->remotepath; # undef because it's not on the server


    # same but file contains "hello"
        my $f = File->new_local("$TMP/temp", "hello");


    # list of files in $DIR (1 level)
        for (File::list($DIR)) {
            say $_->remotepath;
        }


    # list of all files in dir and subdir of $DIR
        File::walk(sub { say $_->remotepath }, File::tree($DIR));

=cut

1;
